/* Copyright 2007,2012 Openstrike
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Version: 0.02
 * Status: Beta
 * Author: Pete Houston
 * Homepage: http://www.apacheconsultancy.com/modules/mod_access_dnsbl/
 *
 * Changelog:
 *   0.02 - Released 16/07/2012
 *      Upgraded to work with Apache 2.4.x: Explicit inclusion of
 *      http_request.h, conditional use of new hooks,
 *      new directive DNSBLAddressType.
 *      Removed declarations of unused variables.
 *      Improved debugging messages.
 *
 *   0.01 - Released 30/05/2007
 *   	First version
 * 
 * Module derived from code originally written by Rob McCool
 * 
 */

#include "apr_strings.h"

#include "http_core.h"
#include "http_request.h"
#include "http_config.h"
#include "http_log.h"

#if APR_HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "dnsbl_lookup.h"

#ifdef APLOG_USE_MODULE
APLOG_USE_MODULE(access_dnsbl);
#endif 

/* If we don't have TRACE levels, just use DEBUG */
#ifndef APLOG_TRACE1
#define APLOG_TRACE1 APLOG_DEBUG
#define APLOG_TRACE2 APLOG_DEBUG
#define APLOG_TRACE3 APLOG_DEBUG
#define APLOG_TRACE4 APLOG_DEBUG
#define APLOG_TRACE5 APLOG_DEBUG
#define APLOG_TRACE6 APLOG_DEBUG
#define APLOG_TRACE7 APLOG_DEBUG
#define APLOG_TRACE8 APLOG_DEBUG
#endif

typedef struct {
    apr_int64_t limited;
    union {
        char *from;
        apr_ipsubnet_t *ip;
    } x;
} allowdeny;

/* things in the 'order' array */
#define DENY_THEN_ALLOW 0
#define ALLOW_THEN_DENY 1
#define MUTUAL_FAILURE  2

/* values for address type > 2.2 only */
#define CLIENT_ADDR     0
#define USERAGENT_ADDR  1

typedef struct {
    int order[METHODS];
    apr_array_header_t *allows;
    apr_array_header_t *denys;
	int addr_type[METHODS];
} access_dir_conf;

module AP_MODULE_DECLARE_DATA access_dnsbl_module;

/* import the lookup function from mod_dnsbl_lookup */
int (*dnsbl_ip)(const char*, int, apr_sockaddr_t*, apr_pool_t*,
        server_rec*, apr_table_t**);

static void dnsbl_funcs(void) {
	dnsbl_ip     = APR_RETRIEVE_OPTIONAL_FN(dnsbl_lookup_ip);
}

static void *create_access_dnsbl_dir_config(apr_pool_t *p, char *dummy)
{
    int i;
    access_dir_conf *conf =
        (access_dir_conf *)apr_pcalloc(p, sizeof(access_dir_conf));

	/* Set the defaults */
    for (i = 0; i < METHODS; ++i) {
        conf->order[i]      = DENY_THEN_ALLOW;
        conf->addr_type[i]  = USERAGENT_ADDR;
    }
    conf->allows = apr_array_make(p, 1, sizeof(allowdeny));
    conf->denys  = apr_array_make(p, 1, sizeof(allowdeny));

    return (void *)conf;
}

static const char *order(cmd_parms *cmd, void *dv, const char *arg)
{
    access_dir_conf *d = (access_dir_conf *) dv;
    int i, o;

    if (!strcasecmp(arg, "allow,deny"))
		o = ALLOW_THEN_DENY;
    else if (!strcasecmp(arg, "deny,allow"))
		o = DENY_THEN_ALLOW;
    else if (!strcasecmp(arg, "mutual-failure"))
		o = MUTUAL_FAILURE;
    else
	return "unknown order";

    for (i = 0; i < METHODS; ++i)
	if (cmd->limited & (AP_METHOD_BIT << i))
	    d->order[i] = o;

    return NULL;
}

static const char *allow_cmd(cmd_parms *cmd, void *dv, const char *from, 
                             const char *where_c)
{
    access_dir_conf *d = (access_dir_conf *) dv;
    allowdeny *a;
    char *where = apr_pstrdup(cmd->pool, where_c);

    if (strcasecmp(from, "from"))
	return "DNSBLAllow and DNSBLDeny must be followed by 'from'";

    a = (allowdeny *) apr_array_push(cmd->info ? d->allows : d->denys);
    a->x.from = where;
    a->limited = cmd->limited;

    return NULL;
}

static const char *addr_type(cmd_parms *cmd, void *dv, const char *arg)
{
    access_dir_conf *d = (access_dir_conf *) dv;
    int i, o;

    if (!strcasecmp(arg, "client"))
		o = CLIENT_ADDR;
    else if (!strcasecmp(arg, "useragent"))
		o = USERAGENT_ADDR;
    else
	return "unknown DNSBLAddressType: should be 'client' or 'useragent'";

    for (i = 0; i < METHODS; ++i)
	if (cmd->limited & (AP_METHOD_BIT << i))
	    d->addr_type[i] = o;

    return NULL;
}

static char its_an_allow;

static const command_rec access_dnsbl_cmds[] =
{
    AP_INIT_TAKE1("dnsblorder", order, NULL, OR_LIMIT,
                  "'allow,deny', 'deny,allow', or 'mutual-failure'"),
    AP_INIT_ITERATE2("dnsblallow", allow_cmd, &its_an_allow, OR_LIMIT,
                     "'from' followed by DNSBLZones"),
    AP_INIT_ITERATE2("dnsbldeny", allow_cmd, NULL, OR_LIMIT,
                     "'from' followed by DNSBLZones"),
    AP_INIT_TAKE1("dnsbladdresstype", addr_type, NULL, OR_LIMIT,
                  "'client' or 'useragent'"),
    {NULL}
};

static int find_allowdeny(request_rec *r, apr_array_header_t *a,
				int method, int addr_type)
{
    allowdeny       *ap    = (allowdeny *) a->elts;
    apr_int64_t     mmask  = (AP_METHOD_BIT << method);
    int             i      = 0;
	char            *ip4   = NULL;
#if AP_MODULE_MAGIC_AT_LEAST(20110724,0)
	ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r,
			"Address Type is %i", addr_type);
	apr_sockaddr_t	*remote_addr = addr_type == CLIENT_ADDR ? r->connection->client_addr : r->useragent_addr;
#else
	apr_sockaddr_t	*remote_addr = r->connection->remote_addr;
#endif

	apr_sockaddr_ip_get(&ip4, r->connection->client_addr);
    ap_log_rerror(APLOG_MARK, APLOG_TRACE3, 0, r,
			"Client address is %s", ip4);
	apr_sockaddr_ip_get(&ip4, r->useragent_addr);
    ap_log_rerror(APLOG_MARK, APLOG_TRACE3, 0, r,
			"Useragent address is %s", ip4);
	apr_sockaddr_ip_get(&ip4, remote_addr);
    ap_log_rerror(APLOG_MARK, APLOG_TRACE3, 0, r,
			"address to test is %s", ip4);
    for (i = 0; i < a->nelts; ++i) {
       	ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r,
			"checking list %i for address %s", i, ip4);
    	if (!(mmask & ap[i].limited))
        	continue;
       	ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r,
			"checking list %i stage 2 for address %s", i, ip4);
		if (dnsbl_ip(ap[i].x.from, DNSBL_ANYPOSTV_RETFIRST, remote_addr, r->pool, r->server, NULL) == DNSBL_POSITIVE) {
        	ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r,
				"client address matched by DNSBL: %s, list %i", ip4, i);
			return 1;
		}
	}

	return 0;
}

static int check_dnsbl_dir_access(request_rec *r)
{
    int method = r->method_number;
    int ret = OK;
    access_dir_conf *a = (access_dir_conf *)
        ap_get_module_config(r->per_dir_config, &access_dnsbl_module);

    if (a->order[method] == ALLOW_THEN_DENY) {
        ret = HTTP_FORBIDDEN;
        if (find_allowdeny(r, a->allows, method, a->addr_type[method]))
            ret = OK;
        if (find_allowdeny(r, a->denys, method, a->addr_type[method]))
            ret = HTTP_FORBIDDEN;
    }
    else if (a->order[method] == DENY_THEN_ALLOW) {
        if (find_allowdeny(r, a->denys, method, a->addr_type[method]))
            ret = HTTP_FORBIDDEN;
        if (find_allowdeny(r, a->allows, method, a->addr_type[method]))
            ret = OK;
    }
    else {
        if (find_allowdeny(r, a->allows, method, a->addr_type[method])
            && !find_allowdeny(r, a->denys, method, a->addr_type[method]))
            ret = OK;
        else
            ret = HTTP_FORBIDDEN;
    }

    if (ret == HTTP_FORBIDDEN
        && (ap_satisfies(r) != SATISFY_ANY || !ap_some_auth_required(r))) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "client denied by DNSBL: %s",
            r->filename);
    } else {
        ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r,
			"client passed by DNSBL: %s",
            r->filename);
	}

    return ret;
}

static void register_hooks(apr_pool_t *p)
{
#if AP_MODULE_MAGIC_AT_LEAST(20110724,0)
    ap_hook_check_access(check_dnsbl_dir_access, NULL, NULL, APR_HOOK_MIDDLE, AP_AUTH_INTERNAL_PER_CONF);
#else
    ap_hook_access_checker(check_dnsbl_dir_access, NULL, NULL, APR_HOOK_MIDDLE);
#endif
	ap_hook_optional_fn_retrieve(dnsbl_funcs, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA access_dnsbl_module =
{
    STANDARD20_MODULE_STUFF,
    create_access_dnsbl_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    access_dnsbl_cmds,
    register_hooks		/* register hooks */
};
